#include <framebuffer.h>
#include <vga.h>
#include <memory.h>
#include <ports.h>
#include <multiboot.h>
#include <stddef.h>
#include <debug.h>


#define VGA_ROW(pos) ((pos)/VGA_COLS)
#define VGA_COL(pos) ((pos)%VGA_COLS)
#define VGA_POS(row, col) ((row)*VGA_COLS + (col))

#define VGA_ADDRESS_PORT 0x3D4
#define VGA_DATA_PORT 0x3D5
#define VGA_REGISTER_CURSOR_POS_LOW 0xF
#define VGA_REGISTER_CURSOR_POS_HIGH 0xE

void *vidmem;

struct vga_cell buffer[VGA_SIZE];

uint64_t cursor;
uint8_t format;

static int vga_setup = 0;
static int vga_type;
gfx_context *fb_context;

void vga_init()
{
    struct fbinfo *fbinfo = kernel_boot_data.fbinfo;
    vga_type = fbinfo->framebuffer_type;
    switch(vga_type)
    {
        case 1:
            fb_context = framebuffer_init(fbinfo);
            break;
        case 2:
            vidmem = VGA_MEMORY;
            memset(vidmem, 0, VGA_SIZE*sizeof(struct vga_cell));
            format = 0x07;
            break;
    }

    vga_setup = 1;
}

static void movecursor()
{
    outb(VGA_ADDRESS_PORT, VGA_REGISTER_CURSOR_POS_LOW);
    outb(VGA_DATA_PORT, (uint8_t)(cursor & 0xFF));
    outb(VGA_ADDRESS_PORT, VGA_REGISTER_CURSOR_POS_HIGH);
    outb(VGA_DATA_PORT, (uint8_t)((cursor >> 8) & 0xFF));
}

static void flush()
{
    if(vga_setup)
        memcpy(vidmem, buffer, sizeof(buffer));
}

static void scroll()
{
    while(cursor >= VGA_SIZE)
    {
        memmove(buffer, &buffer[VGA_POS(1,0)], VGA_COLS*(VGA_ROWS-1)*sizeof(struct vga_cell));
        memset(&buffer[VGA_POS(VGA_ROWS-1, 0)], 0, VGA_COLS*sizeof(struct vga_cell));
        cursor -= VGA_COLS;
    }
}

void vga_write(char c)
{
    switch(c)
    {
        case '\n':
            cursor += VGA_COLS - VGA_COL(cursor);
            break;
        default:
            buffer[cursor++] = (struct vga_cell){.c = c, .f = format};
    }
    scroll();
    switch(vga_type)
    {
        case 1:
            vga2fb(buffer, fb_context, VGA_ROWS, VGA_COLS);
            break;
        case 2:
            flush();
            movecursor();
            break;
    }
}